"""
Reference: https://developers.notion.com/reference/block
"""

from __future__ import annotations
from uuid import UUID
from datetime import datetime
from typing import Union, Optional, Literal, List, Annotated

from pydantic_api.base import BaseModel
from pydantic import Field, HttpUrl, PositiveInt

from ..user import PartialUser
from .rich_text import RichTextObject
from ..parent import PageParentObject, BlockParentObject, DatabaseParentObject
from ..common import (
    IconObject,
    ColorLiteral,
    CodeLanguageLiteral,
)
from ..file import FileObject, _BaseFileObject, _FileExternal, _FileUploaded


BlockTypeLiteral = Literal[
    "bookmark",
    "breadcrumb",
    "bulleted_list_item",
    "callout",
    "child_database",
    "child_page",
    "column",
    "column_list",
    "divider",
    "embed",
    "equation",
    "file",
    "heading_1",
    "heading_2",
    "heading_3",
    "image",
    "link_preview",
    "link_to_page",
    "numbered_list_item",
    "paragraph",
    "pdf",
    "quote",
    "synced_block",
    "table",
    "table_of_contents",
    "table_row",
    "template",
    "to_do",
    "toggle",
    "unsupported",
    "video",
]
"""Reference: https://developers.notion.com/reference/block#keys"""

ParentOfBlock = Union[
    PageParentObject,
    BlockParentObject,
    DatabaseParentObject,
]
"""Databases can be parented by pages, blocks, or by the whole workspace. Reference: https://developers.notion.com/reference/block#parent"""


class BaseBlock(BaseModel):
    """Reference: https://developers.notion.com/reference/block#keys"""

    object: Literal["block"] = "block"
    id: UUID
    parent: ParentOfBlock
    type: BlockTypeLiteral
    created_time: datetime
    created_by: PartialUser
    last_edited_time: datetime
    last_edited_by: PartialUser
    archived: bool
    in_trash: bool
    has_children: bool


class EmptyBlockData(BaseModel):
    pass


# bookmark: Refer to https://developers.notion.com/reference/block#bookmark
class BookMarkBlockData(BaseModel):
    caption: List[RichTextObject] = Field(
        default_factory=list, description="The caption for the bookmark."
    )
    url: str = Field(..., description="The link for the bookmark")


class BookmarkBlock(BaseBlock):
    type: Literal["bookmark"] = "bookmark"
    bookmark: BookMarkBlockData = Field(
        ...,
        description="Bookmark block object",
    )


# breadcrumb: Refer to https://developers.notion.com/reference/block#breadcrumb
class BreadcrumbBlock(BaseBlock):
    type: Literal["breadcrumb"] = "breadcrumb"
    breakdcrumb: EmptyBlockData = Field(
        default_factory=EmptyBlockData,
    )


# bulleted_list_item: Refer to https://developers.notion.com/reference/block#bulleted_list_item
class BulletedListItemBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the bulleted list item."
    )
    color: ColorLiteral
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the bulleted_list_item block.",
    )


class BulletedListItem(BaseModel):
    type: Literal["bulleted_list_item"] = "bulleted_list_item"
    bulleted_list_item: BulletedListItemBlockData = Field(
        ...,
        description="Bulleted list item block object",
    )


# callout: Refer to https://developers.notion.com/reference/block#callout
class CalloutBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the callout."
    )
    icon: Optional[IconObject] = Field(None)
    color: ColorLiteral


class CalloutBlock(BaseBlock):
    type: Literal["callout"] = "callout"
    callout: CalloutBlockData


# child_database: Refer to https://developers.notion.com/reference/block#child_database
class ChildDatabaseBlockData(BaseModel):
    title: str = Field(..., description="The plain text title of the database.")


class ChildDatabaseBlock(BaseBlock):
    """
    Note:

    📘 Creating and updating child_database blocks

    To create or update child_database type blocks, use the Create a database and the Update a database endpoints, specifying the ID of the parent page in the parent body param.
    """

    type: Literal["child_database"] = "child_database"
    child_database: ChildDatabaseBlockData


# child_page: Refer to https://developers.notion.com/reference/block#child_page
class ChildPageBlockData(BaseModel):
    title: str = Field(..., description="The plain text title of the page.")


class ChildPageBlock(BaseBlock):
    """
    Note:

    📘 Creating and updating child_page blocks

    To create or update child_page type blocks, use the Create a page and the Update page endpoints, specifying the ID of the parent page in the parent body param.
    """

    type: Literal["child_page"] = "child_page"
    child_page: ChildPageBlockData


# code: Refer to https://developers.notion.com/reference/block#code
class CodeBlockData(BaseModel):
    caption: List[RichTextObject] = Field(
        default_factory=list,
        description="The rich text in the caption of the code block.",
    )
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the code block."
    )
    language: CodeLanguageLiteral = Field(
        ..., description="The language of the code contained in the code block."
    )


class CodeBlock(BaseBlock):
    type: Literal["code"] = "code"
    code: CodeBlockData


# column: Refer to https://developers.notion.com/reference/block#column_list_and_column
class ColumnBlock(BaseBlock):
    """
    Columns are parent blocks for any block types listed in this reference except for other columns. They do not contain any information within the column property. They can only be appended to column_lists.
    """

    type: Literal["column"] = "column"
    column: EmptyBlockData = Field(
        default_factory=EmptyBlockData,
    )


# column_list: Refer to https://developers.notion.com/reference/block#column_list_and_column
class ColumnListBlock(BaseBlock):
    """
    Column lists are parent blocks for columns. They do not contain any information within the column_list property.

    When creating a column_list block via Append block children, the column_list must have at least two columns, and each column must have at least one child.

    Follow these steps to fetch the content in a column_list:

    Get the column_list ID from a query to Retrieve block children for the parent page.

    Get the column children from a query to Retrieve block children for the column_list.

    Get the content in each individual column from a query to Retrieve block children for the unique column ID.
    """

    type: Literal["column_list"] = "column_list"
    column_list: EmptyBlockData = Field(
        default_factory=EmptyBlockData,
    )


# divider: Refer to https://developers.notion.com/reference/block#divider
class DividerBlock(BaseBlock):
    type: Literal["divider"] = "divider"
    divider: EmptyBlockData = Field(
        default_factory=EmptyBlockData,
    )


# embed: Refer to https://developers.notion.com/reference/block#embed
class EmbedBlockData(BaseModel):
    url: HttpUrl = Field(..., description="The URL of the embed.")


class EmbedBlock(BaseBlock):
    """
    Note:

    🚧 Differences in embed blocks between the Notion app and the API

    The Notion app uses a 3rd-party service, iFramely, to validate and request metadata for embeds given a URL. This works well in a web app because Notion can kick off an asynchronous request for URL information, which might take seconds or longer to complete, and then update the block with the metadata in the UI after receiving a response from iFramely.

    We chose not to call iFramely when creating embed blocks in the API because the API needs to be able to return faster than the UI, and because the response from iFramely could actually cause us to change the block type. This would result in a slow and potentially confusing experience as the block in the response would not match the block sent in the request.

    The result is that embed blocks created via the API may not look exactly like their counterparts created in the Notion app.

    👍 Vimeo video links can be embedded in a Notion page via the public API using the embed block type.

    For supported video sources, see [Supported video types](https://developers.notion.com/reference/block#supported-video-types).
    """

    type: Literal["embed"] = "embed"
    embed: EmbedBlockData


# equation: Refer to https://developers.notion.com/reference/block#equation
class EquationBlockData(BaseModel):
    expression: str = Field(..., description="A KaTeX compatible string.")


class EquationBlock(BaseBlock):
    """
    Equation block objects are represented as children of paragraph blocks. They are nested within a rich text object and contain the following information within the equation property:
    """

    type: Literal["equation"] = "equation"
    equation: EquationBlockData


# file
class BaseBlockFileObject(_BaseFileObject):
    caption: List[RichTextObject] = Field(
        default_factory=list,
        description="The caption for the file block.",
    )


class ExternalBlockFileObject(_BaseFileObject):
    type: Literal["external"] = "external"
    external: _FileExternal


class UploadedBlockFileObject(_BaseFileObject):
    type: Literal["file"] = "file"
    file: _FileUploaded


BlockFileObject = Annotated[
    Union[ExternalBlockFileObject, UploadedBlockFileObject],
    Field(discriminator="type"),
]
"""FileObject + caption field"""


class FileBlock(BaseBlock):
    type: Literal["file"] = "file"
    file: BlockFileObject


# headings: Refer to https://developers.notion.com/reference/block#heading
class HeadingBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the heading."
    )
    color: ColorLiteral
    is_toggleable: bool = Field(
        ...,
        description="Whether or not the heading block is a toggle heading or not. If true, then the heading block toggles and can support children. If false, then the heading block is a static heading block.",
    )


class Heading1Block(BaseBlock):
    type: Literal["heading_1"] = "heading_1"
    heading_1: HeadingBlockData


class Heading2Block(BaseBlock):
    type: Literal["heading_2"] = "heading_2"
    heading_2: HeadingBlockData


class Heading3Block(BaseBlock):
    type: Literal["heading_3"] = "heading_3"
    heading_3: HeadingBlockData


HeadingBlock = Annotated[
    Union[Heading1Block, Heading2Block, Heading3Block],
    Field(discriminator="type"),
]


class HeadingBlockFactory:
    @classmethod
    def new(
        cls,
        rich_text: List[RichTextObject],
        color: ColorLiteral = "default",
        is_toggleable: bool = False,
        heading_type: Literal["heading_1", "heading_2", "heading_3"] = "heading_1",
    ):
        pass


# image: Refer to https://developers.notion.com/reference/block#image
ImageBlockData = FileObject


class ImageBlock(BaseBlock):
    type: Literal["image"] = "image"
    image: ImageBlockData


# link_preview: Refer to https://developers.notion.com/reference/block#link_preview
class LinkPreviewBlockData(BaseModel):
    url: HttpUrl = Field(..., description="The URL of the link preview.")


class LinkPreviewBlock(BaseBlock):
    """
    🚧 The link_preview block can only be returned as part of a response. The API does not support creating or appending link_preview blocks.
    """

    type: Literal["link_preview"] = "link_preview"
    link_preview: LinkPreviewBlockData


# mention: Refer to https://developers.notion.com/reference/block#mention
# mention is not an independent block, it is nested in the rich text object in the paragraph block, so just use MentionObject


# numbered_list_item: Refer to https://developers.notion.com/reference/block#numbered_list_item
class NumberedListItemBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the numbered list item."
    )
    color: ColorLiteral
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the numbered_list_item block.",
    )


class NumberedListItem(BaseModel):
    type: Literal["numbered_list_item"] = "numbered_list_item"
    numbered_list_item: NumberedListItemBlockData = Field(
        ...,
        description="Numbered list item block object",
    )


# paragraph: Refer to https://developers.notion.com/reference/block
class ParagraphBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the paragraph."
    )
    color: ColorLiteral
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the paragraph block.",
    )


class ParagraphBlock(BaseBlock):
    type: Literal["paragraph"] = "paragraph"
    paragraph: ParagraphBlockData


# pdf: Refer to https://developers.notion.com/reference/block#pdf
class PdfBlock(BaseBlock):
    type: Literal["pdf"] = "pdf"
    pdf: BlockFileObject


# quote: Refer to https://developers.notion.com/reference/block#quote
class QuoteBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the quote."
    )
    color: ColorLiteral
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the quote block.",
    )


class QuoteBlock(BaseBlock):
    type: Literal["quote"] = "quote"
    quote: QuoteBlockData


# synced_block: Refer to https://developers.notion.com/reference/block#synced-block
class OriginalSyncedBlockData(BaseModel):
    synced_from: Literal[None] = None
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the synced block.",
    )


class BaseSyncedBlock(BaseBlock):
    type: Literal["synced_block"] = "synced_block"


class OriginalSyncedBlock(BaseSyncedBlock):
    synced_block: OriginalSyncedBlockData


class BlockReference(BaseModel):
    block_id: str


class DuplicateSyncedBlockData(BaseModel):
    synced_from: BlockReference


class DupliateSyncedBlock(BaseSyncedBlock):
    synced_block: DuplicateSyncedBlockData


SyncedBlock = Union[OriginalSyncedBlock, DupliateSyncedBlock]


# table: Refer to https://developers.notion.com/reference/block#table
class TableBlockData(BaseModel):
    table_width: PositiveInt = Field(
        ...,
        description="The number of columns in the table. Note that table_width can only be set when the table is first created.",
    )
    has_column_header: bool = Field(
        ...,
        description="Whether the table has a column header. If true, then the first row in the table appears visually distinct from the other rows.",
    )
    has_row_header: bool = Field(
        ...,
        description="Whether the table has a header row. If true, then the first column in the table appears visually distinct from the other columns.",
    )


class TableBlock(BaseBlock):
    type: Literal["table"] = "table"
    table: TableBlockData


# table_row: Refer to https://developers.notion.com/reference/block#table_rows
class TableRowBlockData(BaseModel):
    cells: List[RichTextObject] = Field(
        default_factory=list,
        description="The cells in the table row.",
    )


class TableRowBlock(BaseBlock):
    type: Literal["table_row"] = "table_row"
    table_row: TableRowBlockData


# table_of_contents: Refer to https://developers.notion.com/reference/block#table_of_contents
class TableOfContentsBlockData(BaseModel):
    color: ColorLiteral


class TableOfContentsBlock(BaseBlock):
    type: Literal["table_of_contents"] = "table_of_contents"
    table_of_contents: TableOfContentsBlockData


# template: Refer to https://developers.notion.com/reference/block#template
class TemplateBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the template."
    )
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the template block.",
    )


class TemplateBlock(BaseBlock):
    """
    ❗️ Deprecation Notice

    As of March 27, 2023 creation of template blocks will no longer be supported.
    """

    type: Literal["template"] = "template"
    template: TemplateBlockData


# to_do: Refer to https://developers.notion.com/reference/block#to_do
class TodoBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The rich text displayed in the To do block."
    )
    checked: Optional[bool] = Field(None, description="Whether the To do is checked.")
    color: ColorLiteral
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the to_do block.",
    )


class TodoBlock(BaseModel):
    type: Literal["to_do"] = "to-do"
    to_do: TodoBlockData


# toggle-blocks: Refer to https://developers.notion.com/reference/block
class ToggleBlockData(BaseModel):
    rich_text: List[RichTextObject] = Field(
        default_factory=list, description="The content of the toggle."
    )
    color: ColorLiteral
    children: List[BlockObject] = Field(
        default_factory=list,
        description="The nested child blocks (if any) of the toggle block.",
    )


class ToggleBlock(BaseBlock):
    type: Literal["toggle"] = "toggle"
    toggle: ToggleBlockData


# video: Refer to https://developers.notion.com/reference/block#video
class VideoBlock(BaseBlock):
    """
    📘 Vimeo video links are not currently supported by the video block type. However, they can be embedded in Notion pages using the embed block type. See Embed for more information.
    """

    type: Literal["video"] = "video"
    video: BlockFileObject


# Union Type
BlockObject = Union[
    BookmarkBlock,
    BreadcrumbBlock,
    BulletedListItem,
    CalloutBlock,
    ChildDatabaseBlock,
    ChildPageBlock,
    CodeBlock,
    ColumnBlock,
    ColumnListBlock,
    DividerBlock,
    EmbedBlock,
    EquationBlock,
    FileBlock,
    Heading1Block,
    Heading2Block,
    Heading3Block,
    ImageBlock,
    LinkPreviewBlock,
    NumberedListItem,
    ParagraphBlock,
    PdfBlock,
    QuoteBlock,
    SyncedBlock,
    TableBlock,
    TableRowBlock,
    TableOfContentsBlock,
    TemplateBlock,
    TodoBlock,
    ToggleBlock,
    VideoBlock,
]

__all__ = [
    "BlockObject",
    "BookmarkBlock",
    "BreadcrumbBlock",
    "BulletedListItem",
    "CalloutBlock",
    "ChildDatabaseBlock",
    "ChildPageBlock",
    "CodeBlock",
    "ColumnBlock",
    "ColumnListBlock",
    "DividerBlock",
    "EmbedBlock",
    "EquationBlock",
    "FileBlock",
    "Heading1Block",
    "Heading2Block",
    "Heading3Block",
    "ImageBlock",
    "LinkPreviewBlock",
    "NumberedListItem",
    "ParagraphBlock",
    "PdfBlock",
    "QuoteBlock",
    "SyncedBlock",
    "TableBlock",
    "TableRowBlock",
    "TableOfContentsBlock",
    "TemplateBlock",
    "TodoBlock",
    "ToggleBlock",
    "VideoBlock",
    "BlockTypeLiteral",
    "ParentOfBlock",
]



from datetime import datetime
from typing import Optional, Literal, Union, Dict, Any, Annotated

from uuid import UUID
from pydantic import Field

from pydantic_api.base import BaseModel
from .user import PartialUser
from .properties import TitleProperty
from .common import IconObject, CoverObject
from .parent import DatabaseParentObject, PageParentObject, WorkspaceParentObject


ParentOfPage = Annotated[
    Union[DatabaseParentObject, PageParentObject, WorkspaceParentObject],
    Field(discriminator="type"),
]
"""
The type of the parent of a page. It could be a `DatabaseParentObject`, `PageParentObject` or `WorkspaceParentObject`. (Reference: https://developers.notion.com/reference/page)

Note: Until Nov 2024, only `DatabaseParentObject` and `PageParentObject` are supported as page's parent. (Reference: https://developers.notion.com/reference/post-page It writes the following as the description of param `parent` when creating a page: 'The parent page or database where the new page is inserted, represented as a JSON object with a page_id or database_id key, and the corresponding ID.')
"""


class Page(BaseModel):
    """Reference: https://developers.notion.com/reference/page"""

    object: Literal["page"] = "page"
    id: UUID
    created_time: datetime
    created_by: PartialUser
    last_edited_time: datetime
    last_edited_by: PartialUser
    archived: bool
    in_trash: bool
    icon: Optional[IconObject] = Field(None)
    cover: Optional[CoverObject] = Field(None)
    properties: Dict[str, Any] = Field(default_factory=dict)
    parent: ParentOfPage
    url: str
    public_url: Optional[str] = Field(None)

    @property
    def title_property(self):
        return TitleProperty.model_validate(self.properties.get("title"))

    @property
    def plain_text_title(self) -> str:
        return "".join(
            [t.plain_text for t in self.title_property.title if t.plain_text]
        )


__all__ = [
    "ParentOfPage",
    "Page",
]



from __future__ import annotations
from typing import Generic, TypeVar, Literal, Optional, Annotated, Union, List, Any

from uuid import UUID
from pydantic import Field, model_validator

from pydantic_api.base import BaseModel

TResult = TypeVar("TResult")


NotionPaginatedDataTypeLiteral = Literal[
    "block",
    "comment",
    "database",
    "page",
    "page_or_database",
    "property_item",
    "user",
]
"""Reference: https://developers.notion.com/reference/intro#pagination"""


class NotionPaginatedData(BaseModel, Generic[TResult]):
    """
    Reference:

    - https://developers.notion.com/reference/intro#pagination
    - https://developers.notion.com/reference/page-property-values#paginated-page-properties
    """

    object: Literal["list"] = "list"
    results: list[TResult]
    next_cursor: Optional[str] = None
    has_more: bool
    type: NotionPaginatedDataTypeLiteral
    request_id: UUID
    next_url: Optional[str] = Field(
        None,
        description="The URL the user can request to get the next page of results.",
        examples=[
            "http://api.notion.com/v1/pages/0e5235bf86aa4efb93aa772cce7eab71/properties/vYdV?start_cursor=LYxaUO&page_size=25"
        ],
    )


StartCursor = Annotated[
    str,
    Field(
        description="If supplied, this endpoint will return a page of results starting after the cursor provided. If not supplied, this endpoint will return the first page of results.",
    ),
]

PageSize = Annotated[
    int,
    Field(
        gt=0,
        le=100,
        description="The number of results to return. The default page size is 100, and the maximum is 100.",
    ),
]


# Sort Object.
# Reference:
# - https://developers.notion.com/reference/post-search
# - https://developers.notion.com/reference/post-database-query-sort
SortDirectionLiteral = Literal["ascending", "descending"]

SortAllowedTimestampLiteral = Literal["last_edited_time", "created_time"]


class BaseSortObject(BaseModel):
    direction: SortDirectionLiteral = Field(
        ...,
        description='The direction to sort. Possible values include "ascending" and "descending".',
    )


class PropertyValueSortObject(BaseSortObject):
    property: str = Field(
        ...,
        description="The name of the property to sort against.",
        examples=["Ingredients"],
    )


class EntryTimestampSortObject(BaseSortObject):
    timestamp: SortAllowedTimestampLiteral = Field(
        ...,
        description='The name of the timestamp to sort against. Possible values include "created_time" and "last_edited_time".',
    )


SortObject = Union[PropertyValueSortObject, EntryTimestampSortObject]


class SortObjectFactory:
    @classmethod
    def new_timestamp_sort(
        cls,
        timestamp: SortAllowedTimestampLiteral,
        direction: SortDirectionLiteral = "ascending",
    ):
        return EntryTimestampSortObject(timestamp=timestamp, direction=direction)

    @classmethod
    def new_property_sort(
        cls,
        property: str,
        direction: SortDirectionLiteral = "ascending",
    ):
        return PropertyValueSortObject(property=property, direction=direction)


# Filter Object
class BaseFilterObject(BaseModel):
    property: str = Field(
        ...,
        description="The name of the property as it appears in the database, or the property ID.",
    )


# Checkbox Filter Object: Refer to https://developers.notion.com/reference/post-database-query-filter#checkbox
class CheckboxFilterObject(BaseFilterObject):
    checkbox: dict[Literal["equals", "does_not_equal"], bool] = Field(
        ..., description="Filters for checkbox values based on equality."
    )

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.checkbox) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Date Filter Object: Refer to https://developers.notion.com/reference/post-database-query-filter#date
class DateFilterObject(BaseFilterObject):
    # TODO: clarify the `Any` here
    date: dict[
        Literal[
            "after",
            "before",
            "equals",
            "is_empty",
            "is_not_empty",
            "next_month",
            "next_week",
            "next_year",
            "on_or_after",
            "on_or_before",
            "past_month",
            "past_week",
            "past_year",
            "this_week",
        ],
        Union[str, Any],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.date) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Files Filter Object
class FilesFilterObject(BaseFilterObject):
    files: dict[Literal["is_empty", "is_not_empty"], bool]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.files) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Formula Filter Object
class FormulaFilterObject(BaseFilterObject):
    formula: Union[
        CheckboxFilterObject, DateFilterObject, NumberFilterObject, RichTextFilterObject
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.formula) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Multi-select Filter Object
class MultiSelectFilterObject(BaseFilterObject):
    # TODO: clarify the `Any` here
    multi_select: dict[
        Literal["contains", "does_not_contain", "is_empty", "is_not_empty"],
        Any,
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.multi_select) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Number Filter Object
class NumberFilterObject(BaseFilterObject):
    # TODO: clarify the `Any` here
    number: dict[
        Literal[
            "equals",
            "does_not_equal",
            "greater_than",
            "greater_than_or_equal_to",
            "less_than",
            "less_than_or_equal_to",
            "is_empty",
            "is_not_empty",
        ],
        Union[bool, float],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.number) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# People Filter Object
class PeopleFilterObject(BaseFilterObject):
    people: dict[
        Literal["contains", "does_not_contain", "is_empty", "is_not_empty"],
        Union[str, bool],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.people) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Relation Filter Object
class RelationFilterObject(BaseFilterObject):
    relation: dict[
        Literal["contains", "does_not_contain", "is_empty", "is_not_empty"],
        Union[str, bool],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.relation) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Rich Text Filter Object
class RichTextFilterObject(BaseFilterObject):
    rich_text: dict[
        Literal[
            "equals",
            "does_not_equal",
            "contains",
            "does_not_contain",
            "starts_with",
            "ends_with",
            "is_empty",
            "is_not_empty",
        ],
        Union[str, bool],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.rich_text) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Rollup Filter Object
class RollupFilterObject(BaseFilterObject):
    # TODO: clarify the `Any` here
    rollup: dict[
        Literal["any", "every", "none"],
        Any,
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.rollup) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Select Filter Object
class SelectFilterObject(BaseFilterObject):
    select: dict[
        Literal["equals", "does_not_equal", "is_empty", "is_not_empty"],
        Union[str, bool],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.select) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Status Filter Object
class StatusFilterObject(BaseFilterObject):
    status: dict[
        Literal["equals", "does_not_equal", "is_empty", "is_not_empty"],
        Union[str, bool],
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.status) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


# Timestamp Filter Object
class BaseTimestampFilterObject(BaseModel):
    timestamp: Literal["created_time", "last_edited_time"]


class CreatedTimeTimestampFilterObject(BaseTimestampFilterObject):
    timestamp: Literal["created_time"] = "created_time"
    created_time: DateFilterObject

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.created_time) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


class LastEditedTimeTimestampFilterObject(BaseTimestampFilterObject):
    timestamp: Literal["last_edited_time"] = "last_edited_time"
    last_edited_time: DateFilterObject

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.last_edited_time) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


TimestampFilterObject = Annotated[
    Union[CreatedTimeTimestampFilterObject, LastEditedTimeTimestampFilterObject],
    Field(discriminator="timestamp"),
]


# Unique ID Filter Object
class UniqueIDFilterObject(BaseFilterObject):
    unique_id: dict[
        Literal[
            "does_not_equal",
            "equals",
            "greater_than",
            "greater_than_or_equal_to",
            "less_than",
            "less_than_or_equal_to",
        ],
        int,
    ]

    @model_validator(mode="after")
    def ensure_exact_one_operand(self):
        if not len(self.unique_id) == 1:
            raise ValueError(
                f"Filter Object can only have exactly 1 operand for a property"
            )
        return self


FilterObject = Union[
    CheckboxFilterObject,
    DateFilterObject,
    FilesFilterObject,
    FormulaFilterObject,
    MultiSelectFilterObject,
    NumberFilterObject,
    PeopleFilterObject,
    RelationFilterObject,
    RichTextFilterObject,
    RollupFilterObject,
    SelectFilterObject,
    StatusFilterObject,
    TimestampFilterObject,
    UniqueIDFilterObject,
]


__all__ = [
    # Pagination Related
    "NotionPaginatedDataTypeLiteral",
    "NotionPaginatedData",
    "StartCursor",
    "PageSize",
    # Sort Object
    "SortObject",
    "SortObjectFactory",
    # Filter Object
    "CheckboxFilterObject",
    "DateFilterObject",
    "FilesFilterObject",
    "FormulaFilterObject",
    "MultiSelectFilterObject",
    "NumberFilterObject",
    "PeopleFilterObject",
    "RelationFilterObject",
    "RichTextFilterObject",
    "RollupFilterObject",
    "SelectFilterObject",
    "StatusFilterObject",
    "TimestampFilterObject",
    "UniqueIDFilterObject",
    "FilterObject",
]



"""
Reference: https://developers.notion.com/reference/parent-object
"""

from typing import Literal, Union, Annotated

from uuid import UUID
from pydantic import Field

from pydantic_api.base import BaseModel


ParentObjectTypeLiteral = Literal["database_id", "page_id", "workspace", "block_id"]


class DatabaseParentObject(BaseModel):
    """Database as a parent."""

    type: Literal["database_id"] = "database_id"
    database_id: UUID


class PageParentObject(BaseModel):
    """Page as a parent."""

    type: Literal["page_id"] = "page_id"
    page_id: UUID


class WorkspaceParentObject(BaseModel):
    """Workspace as a parent. I.e. a root-level page in the workspace."""

    type: Literal["workspace"] = "workspace"
    workspace: Literal[True] = True


class BlockParentObject(BaseModel):
    """Block as a parent."""

    type: Literal["block_id"] = "block_id"
    block_id: UUID


ParentObject = Annotated[
    Union[
        DatabaseParentObject, PageParentObject, WorkspaceParentObject, BlockParentObject
    ],
    Field(discriminator="type"),
]


# The following classes are util classes which are not mentioned in the Notion API documentation.
class ParentObjectFactory:
    @classmethod
    def new_page_parent(cls, page_id: UUID) -> PageParentObject:
        return PageParentObject(page_id=page_id)

    @classmethod
    def new_database_parent(cls, database_id: UUID) -> DatabaseParentObject:
        return DatabaseParentObject(database_id=database_id)

    @classmethod
    def new_workspace_parent(cls) -> WorkspaceParentObject:
        return WorkspaceParentObject()

    @classmethod
    def new_block_parent(cls, block_id: UUID) -> BlockParentObject:
        return BlockParentObject(block_id=block_id)


__all__ = [
    "ParentObjectTypeLiteral",
    "DatabaseParentObject",
    "PageParentObject",
    "WorkspaceParentObject",
    "BlockParentObject",
    "ParentObject",
    "ParentObjectFactory",
]



from typing import Literal, Optional

from pydantic import Field

from pydantic_api.base import BaseModel
from .base import BaseRichTextObject


class TextObject(BaseModel):
    """Text object type."""

    content: str
    link: Optional[dict] = Field(None, description="Link object or null")


class TextRichTextObject(BaseRichTextObject):
    """Rich Text type: Text."""

    type: Literal["text"] = "text"
    text: TextObject

    @classmethod
    def new(cls, content: str, link: Optional[dict] = None):
        return cls(text=TextObject(content=content, link=link))


__all__ = [
    "TextObject",
    "TextRichTextObject",
]



"""
Reference: https://developers.notion.com/reference/property-object
"""

from datetime import datetime
from typing import List, Optional, Literal, Dict, Any

from uuid import UUID
from pydantic import Field, HttpUrl

from pydantic_api.base import BaseModel
from .user import PartialUser
from .block import RichTextObject
from .parent import PageParentObject, BlockParentObject
from .properties import DatabaseProperty
from .common import IconObject, CoverObject


# Database Parent Types
ParentOfDatabase = PageParentObject | BlockParentObject
"""
The type of the parent of a database. It could be a `PageParentObject`, `WorkspaceParent` or `BlockParent`. (Reference: https://developers.notion.com/reference/parent-object)

Note: But for now, the API only support `PageParentObject` as database's parent. (Reference: https://developers.notion.com/reference/create-a-database it says: 'Currently, the parent of a new database must be a Notion page or a wiki database.', note that 'wiki database' is also a page object.)
"""


class Database(BaseModel):
    """Reference: https://developers.notion.com/reference/database"""

    object: Literal["database"] = "database"
    id: UUID
    created_time: datetime
    created_by: PartialUser
    last_edited_time: datetime
    last_edited_by: PartialUser
    title: List[RichTextObject] = Field(default_factory=list)
    description: List[Dict[str, Any]] = Field(default_factory=list)
    icon: Optional[IconObject] = Field(None)
    cover: Optional[CoverObject] = Field(None)
    properties: dict[str, DatabaseProperty]
    parent: ParentOfDatabase
    archived: bool
    url: HttpUrl

    @property
    def plain_text_title(self) -> str:
        return "".join([t.plain_text for t in self.title if t.plain_text])


__all__ = [
    # Database Parent Types
    "ParentOfDatabase",
    # Database
    "Database",
]



from typing import List, Optional

from pydantic import BaseModel, Field

from ..objects import BlockObject
from .base import NotionPaginatedData, StartCursor, PageSize


class AppendBlockChildrenRequest(BaseModel):
    """Reference: https://developers.notion.com/reference/patch-block-children"""

    block_id: str = Field(
        ..., description="Identifier for a block. Also accepts a page ID."
    )
    children: List[BlockObject] = Field(
        ...,
        description="Child content to append to a container block as an array of block objects",
    )
    after: Optional[str] = Field(
        None,
        description="The ID of the existing block that the new block should be appended after.",
    )


AppendBlockChildrenResponse = NotionPaginatedData[BlockObject]
"""Returns a paginated list of newly created first level children block objects. Refer to https://developers.notion.com/reference/patch-block-children"""


class RetrieveBlockRequest(BaseModel):
    """Reference: https://developers.notion.com/reference/retrieve-a-block"""

    block_id: str = Field(..., description="Identifier for a Notion block")


RetrieveBlockResponse = BlockObject
"""Reference: https://developers.notion.com/reference/retrieve-a-block"""


class RetrieveBlockChildrenRequest(BaseModel):
    """Reference: https://developers.notion.com/reference/get-block-children"""

    block_id: str = Field(
        ..., description="Identifier for a Notion block, also accepts a page ID."
    )
    start_cursor: Optional[StartCursor] = None
    page_size: Optional[PageSize] = None


RetrieveBlockChildrenResponse = NotionPaginatedData[BlockObject]
"""Reference: https://developers.notion.com/reference/get-block-children"""

# TODO: UpdateBlockRequest

UpdateBlockResponse = BlockObject
"""Reference: https://developers.notion.com/reference/update-a-block"""


class DeleteBlockRequest(BaseModel):
    """Reference: https://developers.notion.com/reference/delete-a-block"""

    block_id: str = Field(..., description="Identifier for a Notion block")


DeleteBlockResponse = BlockObject
"""Reference: https://developers.notion.com/reference/delete-a-block"""


__all__ = [
    "AppendBlockChildrenRequest",
    "AppendBlockChildrenResponse",
    "RetrieveBlockRequest",
    "RetrieveBlockResponse",
    "RetrieveBlockChildrenRequest",
    "RetrieveBlockChildrenResponse",
    # "UpdateBlockRequest",
    "UpdateBlockResponse",
    "DeleteBlockRequest",
    "DeleteBlockResponse",
]



"""
Reference: https://developers.notion.com/reference/unfurl-attribute-object

A Link Preview is created from an array of unfurl attribute objects.

A Link Preview is a real-time excerpt of authenticated content that unfurls in Notion when an authenticated user shares an enabled link. 
Developers can build Link Preview integrations to customize how links for domains they own look when the links unfurl in a Notion workspace. 
The display of the Link Preview is customizable in terms of content and layout.

TODO: add support for link preview
"""
